/* BMW E9X DPF monitor
 *  
 *  Edgar Delgado Barbosa
 *  20200731
 *  
 *  
 *  CAN BUS OBD2 - E9X
 -------PINOUT-------
 * CAN CS to PD 10
 * CAN SO to PD 11
 * CAN SI to PD 12
 * CAN SCK to PD 13
 * CAN INT to PD 9
*/
#include "mcp_can.h"

#define CAN_BCAST     0x7DF

#define sendPID               1777  //  0x6F1
#define recPID                1554  //  0x612

#define ADRESS_MOTOR_TEMP     2801
#define ADRESS_OIL_TEMP       1112
#define ADRESS_DTSLR          1003
#define ADRESS_REGEN_REQ      1028
#define ADRESS_REGEN_STAT     1450
#define ADRESS_EXHAUS_TEMP    1054
#define ADRESS_ASH            1001
#define ADRESS_SOOT           1002
#define ADRESS_BACK_PRESSURE  1062

byte data[8] = {0x12, 0x04, 0x2C, 0x10, 0x00, 0x00, 0x00, 0x00};       //Dataframe for request

#define CAN_INT 9                                               // Set INT to pin 9
const int SPI_CS_PIN = 10;                                      // Set CS to pin 10
MCP_CAN CAN(SPI_CS_PIN);

long unsigned int rxId;
unsigned char len = 0;
#define REQUEST_TIMEOUT 500
#define LIVE_TIMEOUT 2000
#define READ_INTERVAL 1000
#define READ_ERROR 9999

unsigned char rxBuf[8];
char lcdval[12];
long cTime = 0, cTimeRead = 0, cTimeReadValue, cTimeLive;

float motor_temp, oil_temp, dtslr, regen_req, regen_stat, exhaust_temp, ash, soot, back_pressure;
//---------------------------------------------------------------------------------------------------------------
float getValue(int address, float factor_mul, float factor_add, int n_bytes_result)
{
  cTimeReadValue = millis();

  if (CAN_MSGAVAIL == CAN.checkReceive()) cTimeLive = cTimeReadValue;

  if (cTimeReadValue < (cTimeLive + LIVE_TIMEOUT))
  {
    data[4] = (address & 0xFF00) >> 8;
    data[5] = address & 0xFF;    
    if(CAN.sendMsgBuf(sendPID, 0, 8, data) == CAN_OK)
    {
      rxId = 0;
      while ((rxId != recPID) && (millis() < (cTimeReadValue + REQUEST_TIMEOUT)))
      {
        if (CAN_MSGAVAIL == CAN.checkReceive()) 
        {
          CAN.readMsgBufID(&rxId, &len, rxBuf);
          if(rxId == recPID)
          {
            if (n_bytes_result == 2)
              return (((((long)rxBuf[4] << 8) + rxBuf[5]) * factor_mul) + factor_add);
            else
              return (((((long)rxBuf[4] << 24) + ((long)rxBuf[5] << 16) + ((long)rxBuf[6] << 8) + rxBuf[7]) * factor_mul) + factor_add);
          }
        }
      }
    }
  }
  return (READ_ERROR);
}
//---------------------------------------------------------------------------------------------------------------
void setup() 
{
  Serial.begin(115200);

  // ---------------------------- MCP2515 CAN interface ----------------------------
  while (CAN.begin(CAN_500KBPS, MCP_8MHz) != CAN_OK)    // Initialize MCP2515 running at 8MHz with a baudrate of 500kb/s and the masks and filters disabled.
  {
    Serial.println(F("CAN BUS Shield init fail. Init CAN BUS Shield again."));
    delay(100);
  }
  Serial.println(F("CAN BUS Shield init ok!"));

  CAN.setMode(MODE_NORMAL);                                          // Set operation mode to normal so the MCP2515 sends acks to received data. 
  pinMode(CAN_INT, INPUT);                                           // Configuring pin for /INT input
}
// ================================================================================
void loop() 
{
  cTime = millis();
  if ((cTimeRead + READ_INTERVAL) < cTime)
  {
    cTimeRead = cTime;
  
    // ------------------- Request the values ----------------------
    oil_temp =      getValue(ADRESS_OIL_TEMP, 0.01, -100.0, 2);
    motor_temp =    getValue(ADRESS_MOTOR_TEMP, 0.1, -273.14, 2);
    exhaust_temp =  getValue(ADRESS_EXHAUS_TEMP, 0.031281, -50.0, 2);
    back_pressure = getValue(ADRESS_BACK_PRESSURE, 0.045777, -1000.0, 2);
    soot =          getValue(ADRESS_SOOT, 0.015259, 0.0, 2);
    dtslr =         getValue(ADRESS_DTSLR, 0.001, 0.0, 4);
    regen_req =     getValue(ADRESS_REGEN_REQ, 1.0, 0.0, 2);
    regen_stat =    getValue(ADRESS_REGEN_STAT, 1.0, 0.0, 2);

    dtostrf(motor_temp, 5, 1, lcdval); Serial.print(" motor_temp="); Serial.print(lcdval);
    dtostrf(oil_temp, 5, 1, lcdval); Serial.print(" oil_temp="); Serial.print(lcdval);
    dtostrf(dtslr, 5, 1, lcdval); Serial.print(" dtslr="); Serial.print(lcdval);
    dtostrf(regen_req, 5, 1, lcdval); Serial.print(" regen_req="); Serial.print(lcdval);
    dtostrf(regen_stat, 5, 1, lcdval); Serial.print(" regen_stat="); Serial.print(lcdval);
    dtostrf(exhaust_temp, 5, 1, lcdval); Serial.print(" exhaust_temp="); Serial.print(lcdval);
    dtostrf(ash, 5, 1, lcdval); Serial.print(" ash="); Serial.print(lcdval);
    dtostrf(soot, 5, 1, lcdval); Serial.print(" soot="); Serial.print(lcdval);
    dtostrf(back_pressure, 5, 1, lcdval); Serial.print(" back_pressure="); Serial.print(lcdval);

    if (((int)regen_stat & 0x0F) == 0x01) // ------------ REGEN
    {
      Serial.print("#### REGEN ####"); 
    }
    else if ((((int)regen_req >> 8) & 0x0F) == 0x01)  // ------------ REQUESTED
    {
      Serial.print("REQUESTED"); 
    }
    Serial.print("\n");
    delay(100);
  }
}

// =========================================================================================================================================
// E90
// =========================================================================================================================================
//Motor temperature                                                   84 12 F1 2C 10 0A F1  2801  factor_mul=0.1f       factor_add=-273.14f (ºC)
//Oil temperature                                                     84 12 F1 2C 10 04 58  1112  factor_mul=0.01d      factor_add=-100.0d  (ºC)
//Distance traveled since the last successful regeneration            84 12 F1 2C 10 03 EB  1003  factor_mul=0.001d     factor_add=0.0d     (Km)
//Regeneration requested                                              84 12 F1 2C 10 04 04  1028  factor_mul=1          exactMatch(6)       (Y/N)
//Regeneration status                                                 84 12 F1 2C 10 05 AA  1450  factor_mul=1          maskMatch(2)        (Y/N)
//Exhaust gas temperature upstream of the catalyst - corrected value  84 12 F1 2C 10 04 1E  1054  factor_mul=0.031281f  factor_add=-50.0f   (ºC)
//Ash mass                                                            84 12 F1 2C 10 03 E9  1001  factor_mul=0.015259f  factor_add=0.0f     (g)
//Soot mass                                                           84 12 F1 2C 10 03 EA  1002  factor_mul=0.015259f  factor_add=0.0f     (g)
//Differential pressure                                               84 12 F1 2C 10 04 26  1062  factor_mul=0.045777f  factor_add=-1000.0f (hPa)
